@@ -84,6 +84,123 @@ pub use core::stringify as std_stringify;
 /// ```
 #[macro_export]
 macro_rules! parse {
+    (
+        $callback:path { $($callback_args:tt)* } < $($token:tt)*
+    ) => {
+        $crate::parse_generics_impl! { [$crate::parse_callback] [$callback [$($callback_args)*]] [] [] [$($token)*] }
+    };
+    (
+        $callback:path { $($callback_args:tt)* } $($token:tt)*
+    ) => {
+        $crate::allow_where_clause_impl! { [$crate::parse_callback] [$callback [$($callback_args)*]] [] [$($token)*] }
+    };
+}
+
+/// Callback for `parse_raw` that calls a callback for `parse` 
+#[doc(hidden)]
+#[macro_export]
+macro_rules! parse_callback {
+    (
+        $callback:path
+        [$($callback_args:tt)*]
+        [
+            [$([$([$($g:tt)*])*])?]
+            [$([$([$($r:tt)*])*])?]
+            [$($w:tt)*]
+            $($extra:tt)*
+        ]
+        $($rest:tt)*
+    ) => {
+        $callback ! {
+            $($callback_args)*
+            [ $(< $($($g)*),+ >)? ]
+            [ $(< $($($r)*),+ >)? ]
+            [$($w)*]
+            $($rest)*
+        }
+    };
+}
+
+/// Parses (optional) generics and (optional) subsequent where clause, keeping the structure of the parsed information.
+///
+/// This macro accepts an input in the following form:
+///
+/// ```ignore
+/// $callback_macro { $($callback_macro_args)* }
+/// $(
+///     < $generics >
+///     $( $tokens_between_generics_and_where_clause )*
+///     $(
+///         where $where_clause
+///     )?
+/// )?
+/// $(
+///     $( ; | { $($body)* } )
+///     $($remaining_tokens)*
+/// )?
+/// ```
+///
+/// and expands into
+///
+/// ```ignore
+/// $callback_macro! {
+///     $( $callback_macro_args )*
+///     [
+///        [ $( [ $([ $generics ])* ] )? ]
+///        [ $( [ $([ $generics_without_constraints ]) ] )? ]
+///        [ $( where $where_clause )? ]
+///        $($extra_reserved_for_future_expansion:tt)*
+///     ]
+///     $($( $tokens_between_generics_and_where_clause )*)?
+///     $(
+///         $( ; | { $($body)* } )
+///         $($remaining_tokens)*
+///     )?
+/// }
+/// ```
+///
+/// The $extra_reserved_for_future_expansion ends with multiple placeholders to force users
+/// to use a wildcard match rather than an exact match, allowing to add new data
+/// in future crate versions without breaking compatibility.
+///
+/// # Examples
+///
+/// ```rust
+/// pub trait TheTrait { }
+///
+/// #[doc(hidden)]
+/// pub use generics::parse as generics_parse;
+/// #[doc(hidden)]
+/// pub use std::compile_error as std_compile_error;
+///
+/// #[macro_export]
+/// macro_rules! impl_the_trait {
+///     (
+///         $name:ident $($token:tt)*
+///     ) => {
+///         $crate::generics_parse! {
+///             $crate::impl_the_trait {
+///                 @impl $name
+///             }
+///             $($token)*
+///         }
+///     };
+///     (
+///         @impl $name:ident [[$([$([$($g:tt)*])*])?] [$([$([$($r:tt)*])*])?] [$(where $($w:tt)*)?] $($extra:tt)*]
+///     ) => {
+///         impl $(<$($($g)*),*>)? $crate::TheTrait for $name $(<$($($r)*),*>)? $(where $($w)*)? { }
+///     };
+///     (
+///         @impl $name:ident [[$($g:tt)*] [$($r:tt)*] [$($w:tt)*] $($extra:tt)*] $($token:tt)+ 
+///     ) => {
+///         $crate::std_compile_error!(
+///             "invalid input, allowed input is '$name $( < $generics > $(where $where_clause)? )?'"
+///         );
+///     };
+/// }
+/// ```
+#[macro_export]
+macro_rules! parse_raw {
     (
         $callback:path { $($callback_args:tt)* } < $($token:tt)*
     ) => {
@@ -92,13 +209,27 @@ macro_rules! parse {
     (
         $callback:path { $($callback_args:tt)* } $($token:tt)*
     ) => {
-        $crate::deny_where_clause_impl! { [$callback] [$($callback_args)*] [] [$($token)*] }
+        $crate::allow_where_clause_impl! { [$callback] [$($callback_args)*] [] [$($token)*] }
     };
 }
 
 #[doc(hidden)]
 #[macro_export]
 macro_rules! parse_generics_impl {
+    (
+        [$callback:path]
+        [$($callback_args:tt)*]
+        [$($g:tt)*]
+        [$($r:tt)*]
+        [const $param:ident $($token:tt)*]
+    ) => {
+        $crate::parse_generics_impl! {
+            @param
+            [[const $param] [$param]]
+            [$callback] [$($callback_args)*] [$($g)*] [$($r)*]
+            [$($token)*]
+        }
+    };
     (
         [$callback:path]
         [$($callback_args:tt)*]
@@ -108,7 +239,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! { 
             @param
-            [$param]
+            [[$param] [$param]]
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*] 
             [$($token)*]
         }
@@ -122,11 +253,27 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! { 
             @param
-            [$param]
+            [[$param] [$param]]
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*] 
             [$($token)*]
         }
     };
+    (
+        [$callback:path]
+        [$($callback_args:tt)*]
+        [$($g:tt)*]
+        [$($r:tt)*]
+        [> $($token:tt)*]
+    ) => {
+        $crate::parse_generics_impl! {
+            @done
+            [$callback] [$($callback_args)*]
+            [$($g)*]
+            [$($r)*]
+            []
+            [$($token)*]
+        }
+    };
     (
         [$callback:path]
         [$($callback_args:tt)*]
@@ -155,7 +302,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @param
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$callback:path]
         [$($callback_args:tt)*]
         [$($g:tt)*]
@@ -164,7 +311,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @constrained_param [:]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [] []
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*]
             [$($token)*]
@@ -172,7 +319,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @param
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$callback:path]
         [$($callback_args:tt)*]
         [$($g:tt)*]
@@ -181,7 +328,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @constrained_param [=]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [] []
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*] 
             [$($token)*]
@@ -189,7 +336,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @param
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$callback:path]
         [$($callback_args:tt)*]
         [$($g:tt)*]
@@ -199,15 +346,15 @@ macro_rules! parse_generics_impl {
         $crate::parse_generics_impl! {
             @done
             [$callback] [$($callback_args)*]
-            [$($g)* [$param]]
-            [$($r)* [$param]]
+            [$($g)* [$($gparam)*]]
+            [$($r)* [$($rparam)*]]
             []
             [$($token)*]
         }
     };
     (
         @param
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$callback:path]
         [$($callback_args:tt)*]
         [$($g:tt)*]
@@ -217,15 +364,15 @@ macro_rules! parse_generics_impl {
         $crate::parse_generics_impl! {
             @done
             [$callback] [$($callback_args)*]
-            [$($g)* [$param]]
-            [$($r)* [$param]]
+            [$($g)* [$($gparam)*]]
+            [$($r)* [$($rparam)*]]
             []
             [ > $($token)*]
         }
     };
     (
         @param
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$callback:path]
         [$($callback_args:tt)*]
         [$($g:tt)*]
@@ -235,15 +382,15 @@ macro_rules! parse_generics_impl {
         $crate::parse_generics_impl! {
             @done
             [$callback] [$($callback_args)*]
-            [$($g)* [$param]]
-            [$($r)* [$param]]
+            [$($g)* [$($gparam)*]]
+            [$($r)* [$($rparam)*]]
             []
             [$($token)*]
         }
     };
     (
         @param
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$callback:path]
         [$($callback_args:tt)*]
         [$($g:tt)*]
@@ -253,15 +400,15 @@ macro_rules! parse_generics_impl {
         $crate::parse_generics_impl! {
             @done
             [$callback] [$($callback_args)*]
-            [$($g)* [$param]]
-            [$($r)* [$param]]
+            [$($g)* [$($gparam)*]]
+            [$($r)* [$($rparam)*]]
             []
             [ > $($token)*]
         }
     };
     (
         @param
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$callback:path]
         [$($callback_args:tt)*]
         [$($g:tt)*]
@@ -270,14 +417,14 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             [$callback] [$($callback_args)*]
-            [$($g)* [$param]]
-            [$($r)* [$param]]
+            [$($g)* [$($gparam)*]]
+            [$($r)* [$($rparam)*]]
             [$($token)*]
         }
     };
     (
         @param
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$callback:path]
         [$($callback_args:tt)*]
         [$($g:tt)*]
@@ -292,7 +439,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @param
-        [$param:tt]
+        [[$([$($gparam:tt)*])*] [$($rparam:tt)*]]
         [$callback:path]
         [$($callback_args:tt)*]
         [$($([$($g:tt)*])+)?]
@@ -301,13 +448,13 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::std_compile_error!($crate::std_concat!(
             "missing '>' after '",
-            $crate::std_stringify!( < $($($($g)*),+ ,)? $param ),
+            $crate::std_stringify!( < $($($($g)*),+ ,)? $([$($gparam)*])* ),
             "'"
         ));
     };
     (
         @constrained_param [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$callback:path]
         [$($callback_args:tt)*]
@@ -317,7 +464,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @angles_in_constraint [$kind]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)*] [$($value)*]
             [] []
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*] 
@@ -326,7 +473,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @constrained_param [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$callback:path]
         [$($callback_args:tt)*]
@@ -336,7 +483,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @angles_in_constraint [$kind]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)*] [$($value)*]
             [] []
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*]
@@ -345,7 +492,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @constrained_param [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$callback:path]
         [$($callback_args:tt)*]
@@ -356,15 +503,15 @@ macro_rules! parse_generics_impl {
         $crate::parse_generics_impl! {
             @done
             [$callback] [$($callback_args)*]
-            [$($g)* [$param : $($constraint)*]]
-            [$($r)* [$param]]
+            [$($g)* [$($gparam)* : $($constraint)*]]
+            [$($r)* [$($rparam)*]]
             []
             [$($token)*]
         }
     };
     (
         @constrained_param [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$callback:path]
         [$($callback_args:tt)*]
@@ -375,15 +522,15 @@ macro_rules! parse_generics_impl {
         $crate::parse_generics_impl! {
             @done
             [$callback] [$($callback_args)*]
-            [$($g)* [$param : $($constraint)*]]
-            [$($r)* [$param]]
+            [$($g)* [$($gparam)* : $($constraint)*]]
+            [$($r)* [$($rparam)*]]
             []
             [ > $($token)*]
         }
     };
     (
         @constrained_param [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$callback:path]
         [$($callback_args:tt)*]
@@ -394,15 +541,15 @@ macro_rules! parse_generics_impl {
         $crate::parse_generics_impl! {
             @done
             [$callback] [$($callback_args)*]
-            [$($g)* [$param : $($constraint)*]]
-            [$($r)* [$param]]
+            [$($g)* [$($gparam)* : $($constraint)*]]
+            [$($r)* [$($rparam)*]]
             []
             [$($token)*]
         }
     };
     (
         @constrained_param [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$callback:path]
         [$($callback_args:tt)*]
@@ -413,15 +560,15 @@ macro_rules! parse_generics_impl {
         $crate::parse_generics_impl! {
             @done
             [$callback] [$($callback_args)*]
-            [$($g)* [$param : $($constraint)*]]
-            [$($r)* [$param]]
+            [$($g)* [$($gparam)* : $($constraint)*]]
+            [$($r)* [$($rparam)*]]
             []
             [ > $($token)*]
         }
     };
     (
         @constrained_param [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$callback:path]
         [$($callback_args:tt)*]
@@ -431,14 +578,14 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             [$callback] [$($callback_args)*]
-            [$($g)* [$param : $($constraint)*]]
-            [$($r)* [$param]]
+            [$($g)* [$($gparam)* : $($constraint)*]]
+            [$($r)* [$($rparam)*]]
             [$($token)*]
         }
     };
     (
         @constrained_param [:]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$callback:path]
         [$($callback_args:tt)*]
@@ -448,7 +595,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @constrained_param [=]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)*] [$($value)*]
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*]
             [$($token)*]
@@ -456,7 +603,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @constrained_param [:]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$callback:path]
         [$($callback_args:tt)*]
@@ -466,7 +613,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @constrained_param [:]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)* $x] [$($value)*]
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*]
             [$($token)*]
@@ -474,7 +621,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @constrained_param [=]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$callback:path]
         [$($callback_args:tt)*]
@@ -484,7 +631,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @constrained_param [=]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)*] [$($value)* $x]
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*]
             [$($token)*]
@@ -492,7 +639,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @constrained_param [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($($constraint:tt)+)?] [$($($value:tt)+)?]
         [$callback:path]
         [$($callback_args:tt)*]
@@ -502,13 +649,13 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::std_compile_error!($crate::std_concat!(
             "missing '>' after '",
-            $crate::std_stringify!( < $($($($g)*),+ ,)? $param $( : $($constraint)+)? $( = $($value)+)? ),
+            $crate::std_stringify!( < $($($($g)*),+ ,)? $($gparam)* $( : $($constraint)+)? $( = $($value)+)? ),
             "'"
         ));
     };
     (
         @angles_in_constraint [:]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$($inside_angles:tt)*]
         []
@@ -520,7 +667,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @constrained_param [:]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)* < $($inside_angles)* > ] [$($value)*]
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*]
             [$($token)*]
@@ -528,7 +675,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @angles_in_constraint [=]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$($inside_angles:tt)*]
         []
@@ -540,7 +687,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @constrained_param [=]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)*] [$($value)* < $($inside_angles)* > ]
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*]
             [$($token)*]
@@ -548,7 +695,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @angles_in_constraint [:]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$($inside_angles:tt)*]
         []
@@ -560,7 +707,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @constrained_param [:]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)* < $($inside_angles)* > ] [$($value)*]
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*]
             [ > $($token)*]
@@ -568,7 +715,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @angles_in_constraint [=]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$($inside_angles:tt)*]
         []
@@ -580,7 +727,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @constrained_param [=]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)*] [$($value)* < $($inside_angles)* > ]
             [$callback] [$($callback_args)*] [$($g)*] [$($r)*]
             [ > $($token)*]
@@ -588,7 +735,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @angles_in_constraint [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$($inside_angles:tt)*]
         [[$($parent_level:tt)*] $([$($outer_levels:tt)*])*]
@@ -600,7 +747,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @angles_in_constraint [$kind]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)*] [$($value)*]
             [$($parent_level)* < $($inside_angles)* > ]
             [$([$($outer_levels)*])*]
@@ -610,7 +757,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @angles_in_constraint [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$($inside_angles:tt)*]
         [[$($parent_level:tt)*] $([$($outer_levels:tt)*])*]
@@ -622,7 +769,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @angles_in_constraint [$kind]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)*] [$($value)*]
             [$($parent_level)* < $($inside_angles)* > ]
             [$([$($outer_levels)*])*]
@@ -632,7 +779,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @angles_in_constraint [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$($inside_angles:tt)*]
         [$([$($outer_levels:tt)*])*]
@@ -644,7 +791,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @angles_in_constraint [$kind]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)*] [$($value)*]
             []
             [[$($inside_angles:tt)*] $([$($outer_levels)*])*]
@@ -654,7 +801,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @angles_in_constraint [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$($inside_angles:tt)*]
         [$([$($outer_levels:tt)*])*]
@@ -666,7 +813,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @angles_in_constraint [$kind]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)*] [$($value)*]
             []
             [[$($inside_angles:tt)*] $([$($outer_levels)*])*]
@@ -676,7 +823,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @angles_in_constraint [$kind:tt]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($value:tt)*]
         [$($inside_angles:tt)*]
         [$([$($outer_levels:tt)*])*]
@@ -688,7 +835,7 @@ macro_rules! parse_generics_impl {
     ) => {
         $crate::parse_generics_impl! {
             @angles_in_constraint [$kind]
-            [$param]
+            [[$($gparam)*] [$($rparam)*]]
             [$($constraint)*] [$($value)*]
             [$($inside_angles)* $x]
             [$([$($outer_levels)*])*]
@@ -698,7 +845,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @angles_in_constraint [:]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($constraint:tt)*] [$($($value:tt)+)?]
         [$($inside_angles:tt)*]
         [$([$($outer_levels:tt)*])*]
@@ -711,7 +858,7 @@ macro_rules! parse_generics_impl {
         $crate::std_compile_error!($crate::std_concat!(
             "missing '>' after '",
             $crate::std_stringify!(
-                < $($($($g)*),+ ,)? $param
+                < $($($($g)*),+ ,)? $($gparam)*
                 : $($constraint)* $( < $($outer_levels)* )* < $($inside_angles)*
                 $( = $($value)+)?
             ),
@@ -720,7 +867,7 @@ macro_rules! parse_generics_impl {
     };
     (
         @angles_in_constraint [=]
-        [$param:tt]
+        [[$($gparam:tt)*] [$($rparam:tt)*]]
         [$($($constraint:tt)+)?] [$($value:tt)*]
         [$($inside_angles:tt)*]
         [$([$($outer_levels:tt)*])*]
@@ -733,7 +880,7 @@ macro_rules! parse_generics_impl {
         $crate::std_compile_error!($crate::std_concat!(
             "missing '>' after '",
             $crate::std_stringify!(
-                < $($($($g)*),+ ,)? $param
+                < $($($($g)*),+ ,)? $($gparam)*
                 $( : $($constraint)+)?
                 = $($value)* $( < $($outer_levels)* )* < $($inside_angles)*
             ),
@@ -744,16 +891,19 @@ macro_rules! parse_generics_impl {
         @done
         [$callback:path]
         [$($callback_args:tt)*]
-        [$([$($g:tt)*])+]
-        [$([$($r:tt)*])+]
+        [$($g:tt)*]
+        [$($r:tt)*]
         [$($inter:tt)*]
         [ ; $($token:tt)*]
     ) => {
         $callback ! {
             $($callback_args)*
-            [ < $($($g)*),+ > ]
-            [ < $($($r)*),+ > ]
-            []
+            [
+                [ [$($g)*] ]
+                [ [$($r)*] ]
+                []
+                $crate $crate
+            ]
             $($inter)* ; $($token)*
         }
     };
@@ -761,16 +911,19 @@ macro_rules! parse_generics_impl {
         @done
         [$callback:path]
         [$($callback_args:tt)*]
-        [$([$($g:tt)*])+]
-        [$([$($r:tt)*])+]
+        [$($g:tt)*]
+        [$($r:tt)*]
         [$($inter:tt)*]
         [ $( { $($body:tt)* } $($token:tt)* )? ]
     ) => {
         $callback ! {
             $($callback_args)*
-            [ < $($($g)*),+ > ]
-            [ < $($($r)*),+ > ]
-            []
+            [
+                [ [$($g)*] ]
+                [ [$($r)*] ]
+                []
+                $crate $crate
+            ]
             $($inter)* $( { $($body)* } $($token)* )?
         }
     };
@@ -778,16 +931,16 @@ macro_rules! parse_generics_impl {
         @done
         [$callback:path]
         [$($callback_args:tt)*]
-        [$([$($g:tt)*])+]
-        [$([$($r:tt)*])+]
+        [$($g:tt)*]
+        [$($r:tt)*]
         [$($inter:tt)*]
         [where $($token:tt)*]
     ) => {
         $crate::parse_where_clause_impl! {
             [$callback]
             [$($callback_args)*]
-            [ < $($($g)*),+ > ]
-            [ < $($($r)*),+ > ]
+            [ [$($g)*] ]
+            [ [$($r)*] ]
             [] [$($inter)*] [$($token)*]
         }
     };
@@ -795,16 +948,16 @@ macro_rules! parse_generics_impl {
         @done
         [$callback:path]
         [$($callback_args:tt)*]
-        [$($g:tt)+]
-        [$($r:tt)+]
+        [$($g:tt)*]
+        [$($r:tt)*]
         [$($inter:tt)*]
         [$token:tt $($other_tokens:tt)*]
     ) => {
         $crate::parse_generics_impl! {
             @done
             [$callback] [$($callback_args)*]
-            [$($g)+]
-            [$($r)+]
+            [$($g)*]
+            [$($r)*]
             [$($inter)* $token]
             [$($other_tokens)*]
         }
@@ -824,9 +977,12 @@ macro_rules! parse_where_clause_impl {
     ) => {
         $callback ! { 
             $($callback_args)*
-            [$($g)*]
-            [$($r)*]
-            [$(where $($w)+)?]
+            [
+                [$($g)*]
+                [$($r)*]
+                [$(where $($w)+)?]
+                $crate $crate
+            ]
             $($inter)* ; $($token)*
         }
     };
@@ -840,9 +996,12 @@ macro_rules! parse_where_clause_impl {
     ) => {
         $callback ! { 
             $($callback_args)*
-            [$($g)*]
-            [$($r)*]
-            [$(where $($w)+)?]
+            [
+                [$($g)*]
+                [$($r)*]
+                [$(where $($w)+)?]
+                $crate $crate
+            ]
             $($inter)* $( { $($body)* } $($token)* )?
         }
     };
@@ -867,7 +1026,7 @@ macro_rules! parse_where_clause_impl {
 
 #[doc(hidden)]
 #[macro_export]
-macro_rules! deny_where_clause_impl {
+macro_rules! allow_where_clause_impl {
     (
         [$callback:path]
         [$($callback_args:tt)*]
@@ -876,9 +1035,12 @@ macro_rules! deny_where_clause_impl {
     ) => {
         $callback ! {
             $($callback_args)*
-            []
-            []
-            []
+            [
+                []
+                []
+                []
+                $crate $crate
+            ]
             $($inter)* ; $($token)*
         }
     };
@@ -890,9 +1052,12 @@ macro_rules! deny_where_clause_impl {
     ) => {
         $callback ! {
             $($callback_args)*
-            []
-            []
-            []
+            [
+                []
+                []
+                []
+                $crate $crate
+            ]
             $($inter)* $( { $($body)* } $($token)* )?
         }
     };
@@ -902,7 +1067,14 @@ macro_rules! deny_where_clause_impl {
         [$($inter:tt)*]
         [where $($token:tt)*]
     ) => {
-        $crate::std_compile_error!("unexpected 'where' without generics preceding");
+        $crate::parse_where_clause_impl! { 
+            [$callback]
+            [$($callback_args)*]
+            [] []
+            []
+            [$($inter)*]
+            [$($token)*]
+        }
     };
     (
         [$callback:path]
@@ -910,7 +1082,7 @@ macro_rules! deny_where_clause_impl {
         [$($inter:tt)*]
         [$token:tt $($other_tokens:tt)*]
     ) => {
-        $crate::deny_where_clause_impl! {
+        $crate::allow_where_clause_impl! {
             [$callback] [$($callback_args)*]
             [$($inter)* $token]
             [$($other_tokens)*]
@@ -1020,7 +1192,7 @@ macro_rules! concat_g_impl {
     ) => {
         $crate::std_compile_error!($crate::std_concat!(
             "invalid generics '",
-            $crate::std_stringify!($($item)*),
+            $crate::std_stringify!($($item:tt)*),
             "'"
         ));
     };
@@ -1165,7 +1337,7 @@ macro_rules! concat_g_impl {
     ) => {
         $crate::std_compile_error!($crate::std_concat!(
             "invalid generics '",
-            $crate::std_stringify!($($tail)+),
+            $crate::std_stringify!($($tail:tt)+),
             "'"
         ));
     };
@@ -1400,7 +1572,7 @@ macro_rules! concat_r_impl {
     ) => {
         $crate::std_compile_error!($crate::std_concat!(
             "invalid generics '",
-            $crate::std_stringify!($($tail)+),
+            $crate::std_stringify!($($tail:tt)+),
             "'"
         ));
     };
@@ -1457,7 +1629,7 @@ macro_rules! concat_w_impl {
     ) => {
         $crate::std_compile_error!($crate::std_concat!(
             "invalid generics '",
-            $crate::std_stringify!($($item)+),
+            $crate::std_stringify!($($item:tt)+),
             "'"
         ));
     };
@@ -1664,36 +1836,4 @@ mod tests {
         let _ = x.c;
         let _ = x.d;
     }
-
-    macro_rules! descr {
-        (
-            impl $($token:tt)*
-        ) => {
-            parse! {
-                descr {
-                    @impl
-                }
-                $($token)*
-            }
-        };
-        (
-            @impl [$($g:tt)*] [$($r:tt)*] [$($w:tt)*] $t:ty { $($body:tt)* }
-        ) => {
-            impl $($g)* $t $($w)* { $($body)* }
-        };
-    }
-
-    struct Test<A>(A);
-
-    descr! {
-        impl<A> Test<A> where A: Clone {
-            fn a(&self) { }
-        }
-    }
-
-    #[test]
-    fn run_test() {
-        let x: Test<u32> = Test(7);
-        x.a();
-    }
 }
